@@ -49,6 +49,8 @@
 
 /* Commands from the popup menu */
 enum {
+	POPUP_ROTATE_CLOCKWISE,
+	POPUP_ROTATE_COUNTER_CLOCKWISE,
 	POPUP_ZOOM_IN,
 	POPUP_ZOOM_OUT,
 	POPUP_ZOOM_1,
@@ -769,13 +771,23 @@ popup_menu_cb (gpointer data, guint action, GtkWidget *widget)
 	case POPUP_CLOSE:
 		g_signal_emit (image_view, signals[CLOSE_ITEM_ACTIVATED], 0);
 		break;
-
+	case POPUP_ROTATE_COUNTER_CLOCKWISE:
+		apply_transformation (EOG_IMAGE_VIEW (data), eog_transform_rotate_new (270));
+		break;
+	case POPUP_ROTATE_CLOCKWISE:
+		apply_transformation (EOG_IMAGE_VIEW (data), eog_transform_rotate_new (90));
+		break;
 	default:
 		g_assert_not_reached ();
 	}
 }
 
 static GtkItemFactoryEntry popup_entries[] = {
+	{ N_("/Rotate C_lockwise"), NULL, popup_menu_cb, POPUP_ROTATE_CLOCKWISE,
+	  "<Item>", NULL },
+	{ N_("/Rotate Counte_r Clockwise"), NULL, popup_menu_cb, POPUP_ROTATE_COUNTER_CLOCKWISE,
+	  "<Item>", NULL },
+	{ "/sep", NULL, NULL, 0, "<Separator>", NULL },
 	{ N_("/_Zoom In"), NULL, popup_menu_cb, POPUP_ZOOM_IN,
 	  "<StockItem>", GTK_STOCK_ZOOM_IN },
 	{ N_("/Zoom _Out"), NULL, popup_menu_cb, POPUP_ZOOM_OUT,
@@ -803,7 +815,7 @@ static void
 setup_item_factory (EogImageView *image_view, gboolean need_close_item)
 {
 	EogImageViewPrivate *priv;
-
+	
 	priv = image_view->priv;
 
 	priv->item_factory = gtk_item_factory_new (GTK_TYPE_MENU, "<main>", NULL);
@@ -870,8 +882,6 @@ image_size_prepared_cb (EogImage *img, int width, int height, gpointer data)
 	view = EOG_IMAGE_VIEW (data);
 	priv = view->priv;
 
-	g_print ("size prepared cb\n");
-
 	arg = bonobo_arg_new (BONOBO_ARG_INT);
 
 	eog_image_view_get_prop (NULL, arg, PROP_WINDOW_WIDTH, NULL, view);
@@ -907,6 +917,27 @@ image_progress_cb (EogImage *img, float progress, gpointer data)
 	bonobo_arg_release (arg);
 }
 
+static void
+image_changed_cb (EogImage *img, gpointer data)
+{
+	EogImageView *view;
+	BonoboArg *arg;
+
+	g_return_if_fail (EOG_IS_IMAGE_VIEW (data));
+	g_return_if_fail (EOG_IS_IMAGE (img));
+	
+	view = EOG_IMAGE_VIEW (data);
+
+	/* update interested window status listeners */
+	arg = bonobo_arg_new (BONOBO_ARG_STRING);
+	eog_image_view_get_prop (NULL, arg, PROP_WINDOW_STATUS, NULL, view);
+
+	bonobo_event_source_notify_listeners (view->priv->property_bag->es,
+					      "window/status",
+					      arg, NULL);
+	bonobo_arg_release (arg);
+}
+
 static gint
 load_uri_cb (BonoboPersistFile *pf, const CORBA_char *text_uri,
 	     CORBA_Environment *ev, void *closure)
@@ -923,6 +954,7 @@ load_uri_cb (BonoboPersistFile *pf, const CORBA_char *text_uri,
 	image = eog_image_new (text_uri);
 	g_signal_connect (image, "loading_size_prepared", G_CALLBACK (image_size_prepared_cb), view);
 	g_signal_connect (image, "progress", G_CALLBACK (image_progress_cb), view);
+	g_signal_connect (image, "image_changed", G_CALLBACK (image_changed_cb), view);
 
 	/* FIXME: remove potential signal handlers from old image object */
 	if (priv->image != NULL) {
